@@ -12,6 +12,7 @@
 #include "edit_text.hpp"
 #include "file.hpp"
 #include "analyze.hpp"
+#include "scheme.hpp"
 
 /******************************************************************************
 * Constructors and destructors
@@ -28,7 +29,7 @@ void
 edit_text_rep::correct_concat (path p, int done) {
   tree t (subtree (et, p));
   if (L(t) != CONCAT) {
-    cerr << "\nThe tree was <" << t << ">\n";
+    failed_error << "The tree was '" << t << "'\n";
     FAILED ("concat expected");
   }
 
@@ -60,19 +61,19 @@ edit_text_rep::correct_concat (path p, int done) {
       return;
     }
     else if (is_multi_paragraph (t[i]) &&
-	     is_document (subtree (et, path_up (p))))
+             is_document (subtree (et, path_up (p))))
       {
-	if ((i+1)<n) {
-	  split (p * (i+1));
-	  correct_concat (path_inc (p));
-	}
-	if (i==0) remove_node (p * 0);
-	else {
-	  split (p * i);
-	  remove_node (path_inc (p) * 0);
-	  correct_concat (p);
-	}
-	return;
+        if ((i+1)<n) {
+          split (p * (i+1));
+          correct_concat (path_inc (p));
+        }
+        if (i==0) remove_node (p * 0);
+        else {
+          split (p * i);
+          remove_node (path_inc (p) * 0);
+          correct_concat (p);
+        }
+        return;
       }
   }
 }
@@ -95,7 +96,9 @@ edit_text_rep::pure_line (path p) {
   return
     is_document (st) ||
     ((is_func (st, WITH) || is_func (st, LOCUS) ||
-      is_func (st, CANVAS) || is_func (st, ORNAMENT)) &&
+      is_func (st, CANVAS) ||
+      is_func (st, ORNAMENT) || is_func (st, ART_BOX) ||
+      is_func (st, WITH_PACKAGE)) &&
      (last_item (p) == (N(st)-1)) && pure_line (p)) ||
     (is_extension (st) && (last_item (p) >= 0) && pure_line (p));
 }
@@ -113,9 +116,11 @@ edit_text_rep::accepts_return (path p) {
     (is_func (st, XMACRO, 2) && (last_item (p) == 1)) ||
     ((is_func (st, WITH) || is_mod_active (st) ||
       is_func (st, STYLE_WITH) || is_func (st, VAR_STYLE_WITH) ||
-      is_func (st, LOCUS) ||
-      is_func (st, CANVAS) || is_func (st, ORNAMENT)) &&
+      is_func (st, LOCUS) || is_func (st, WITH_PACKAGE) ||
+      is_func (st, CANVAS)) &&
      (last_item (p) == (N(st)-1)) && pure_line (p)) ||
+    ((is_func (st, ORNAMENT) || is_func (st, ART_BOX)) &&
+     (last_item (p) == 0) && pure_line (p)) ||
     (is_extension (st) && (last_item (p) >= 0) && pure_line (p));
 }
 
@@ -169,12 +174,12 @@ edit_text_rep::prepare_for_insert () {
       insert_node (p * 0, DOCUMENT);
     else {
       if (l == 0) {
-	insert (p, tree (DOCUMENT, ""));
-	go_to (p * 0);
+        insert (p, tree (DOCUMENT, ""));
+        go_to (p * 0);
       }
       else {
-	insert (path_inc (p), tree (DOCUMENT, ""));
-	go_to (path_inc (p) * 0);
+        insert (path_inc (p), tree (DOCUMENT, ""));
+        go_to (path_inc (p) * 0);
       }
     }
     return prepare_for_insert ();
@@ -197,10 +202,15 @@ edit_text_rep::prepare_for_insert () {
 
 void
 edit_text_rep::insert_tree (tree t, path p_in_t) {
+  if (!as_bool (call ("like-emacs?"))) selection_cut ("none");
   if (is_atomic (t) && (p_in_t == end (t)) &&
       is_atomic (subtree (et, path_up (tp))))
     insert (tp, t);
   else if (is_document (t)) {
+    if (subtree (et, path_up (tp)) == "" &&
+        accepts_return (path_up (tp)) &&
+        !is_func (subtree (et, path_up (tp, 2)), DOCUMENT))
+      insert_node (path_up (tp) * 0, DOCUMENT);
     if (insert_return ()) return;
     path p= search_parent_upwards (DOCUMENT);
     bool empty= (subtree (et, p) == "");
@@ -212,6 +222,10 @@ edit_text_rep::insert_tree (tree t, path p_in_t) {
     if (!empty) remove_return (path_add (p, N(t)-2));
   }
   else if (is_multi_paragraph (t)) {
+    if (subtree (et, path_up (tp)) == "" &&
+        accepts_return (path_up (tp)) &&
+        !is_func (subtree (et, path_up (tp, 2)), DOCUMENT))
+      insert_node (path_up (tp) * 0, DOCUMENT);
     if (make_return_after ()) return;
     path p= search_parent_upwards (DOCUMENT);
     if (subtree (et, p) == "") remove (p, 1);
@@ -253,7 +267,7 @@ get_unit (tree t) {
   int i;
   string s= as_string (t);
   for (i=0; i<N(s); i++)
-    if ((s[i]>='a') && (s[i]<='z')) break;
+    if (is_locase (s[i])) break;
   return s (i, N(s));
 }
 
@@ -262,7 +276,7 @@ get_quantity (tree t) {
   int i;
   string s= as_string (t);
   for (i=0; i<N(s); i++)
-    if ((s[i]>='a') && (s[i]<='z')) break;
+    if (is_locase (s[i])) break;
   return s (0, i);
 }
 
@@ -346,19 +360,17 @@ void
 edit_text_rep::make_image (
   string file_name, bool link, string w, string h, string x, string y)
 {
-  url image= url_system (file_name);
+  url image= url_unix (file_name);
   string type= "";
-  if (is_rooted (image))
-    image= delta (get_name (), image);
-
   tree t (IMAGE);
   if (link) {
     if (is_rooted (image, "default")) image= reroot (image, "file");
-    t << as_string (image, URL_STANDARD);
+    // here we convert the filename to cork-universal because it is exposed in the document, in the <image> tag
+    t << utf8_to_cork (as_string (image, URL_STANDARD));
   }
   else {
     string s;
-    tree vim= verbatim (as_string (image));
+    tree vim= verbatim (utf8_to_cork (as_string (image)));
     load_string (relative (get_name (), image), s, false);
     if (s == "") {
       set_message (concat ("File '", vim, "' not found"), "make image");
@@ -369,7 +381,7 @@ edit_text_rep::make_image (
       set_message (concat ("File '", vim, "' is not an image"), "make image");
       return;
     }
-    t << tuple (tree (RAW_DATA, s), type);
+    t << tuple (tree (RAW_DATA, s), as_string (tail (image)));
   }
   t << tree (w) << tree (h) << tree (x) << tree (y);
   insert_tree (t);
