@@ -20,12 +20,12 @@
 #include "QTMMenuHelper.hpp"
 #include "QTMGuiHelper.hpp"
 #include "QTMStyle.hpp"
-#include "QTMApplication.hpp"
 #include "QTMTreeModel.hpp"
 
 #include <QToolTip>
 #include <QCompleter>
 #include <QKeyEvent>
+#include <QApplication>
 
 /******************************************************************************
  * QTMCommand
@@ -34,19 +34,20 @@
 /*! Queues the object's command into the main queue. */
 void 
 QTMCommand::apply()  {
+BEGIN_SLOT
   if (!is_nil (cmd)) {
     the_gui->process_command (cmd);
     if (DEBUG_QT) {
       debug_qt << "QTMCommand::apply() (delayed)\n";
       /* FIXME: this sometimes crashes:
-      cmd->print(debug_qt);
-      debug_qt << "\n";
-       */
+         cmd->print(debug_qt);
+         debug_qt << "\n";
+      */
     }
   }
+END_SLOT
 }
 
-
 /******************************************************************************
  * QTMAction
  ******************************************************************************/
@@ -67,8 +68,10 @@ void
 QTMAction::set_text (string s) {
   if (N(s)) {
       // FIXME: this will only work if the system language is English!
-    if (s == "Help" || s == "Edit" || s == "Preferences...")
+    if (s == "Help" || s == "Edit" || s == "View" ||
+        s == "Preferences...")
       s = s * " ";
+    s= replace (s, "&", "&&");
     str = s;
     setText (to_qstring (s));
   }
@@ -76,14 +79,17 @@ QTMAction::set_text (string s) {
 
 void 
 QTMAction::doRefresh() {
+BEGIN_SLOT
   set_text (str);
+END_SLOT
 }
 
 void
-QTMAction::showToolTip()
-{
+QTMAction::showToolTip() {
+BEGIN_SLOT
   _timer->start (500);   // Restarts the timer if already running
   _pos = QCursor::pos();
+END_SLOT
 }
 
 /*
@@ -98,17 +104,19 @@ QTMAction::showToolTip()
  stops, the distance is bigger than the given constant and no tooltip is
  displayed.
  */
+
 void
 QTMAction::doShowToolTip() {
+BEGIN_SLOT
   static int step = QApplication::font().pointSize();
   _timer->stop();
   if ((QCursor::pos() - _pos).manhattanLength() < step)  // Hideous HACK
     QToolTip::showText (QCursor::pos(), toolTip());
   else
     QToolTip::hideText();
+END_SLOT
 }
 
-
 /******************************************************************************
  * QTMWidgetAction
  ******************************************************************************/
@@ -143,13 +151,6 @@ QTMTileAction::QTMTileAction (array<widget>& arr, int _cols, QObject* parent)
   };
 }
 
-/*!
- FIXME: QTMTileAction::createWidget is called twice:
- the first time when the action is added to the menu,
- the second when from the menu it is transferred to the toolbar.
- This is weird since the first widget does not ever use
- the widget so it results in a waste of time.
- */
 QWidget*
 QTMTileAction::createWidget (QWidget* parent)
 {
@@ -189,18 +190,12 @@ QTMMinibarAction::QTMMinibarAction (array<widget>& arr, QObject* parent)
   actions.reserve (N (arr));
   for (int i = 0; i < N (arr); i++) {
     if (is_nil (arr[i])) break;
-    QAction *act = concrete (arr[i])->as_qaction();
+    QAction* act = concrete (arr[i])->as_qaction();
     act->setParent (this);
     actions.append (act);
   };
 }
 
-/*!
- FIXME: QTMMinibarAction::createWidget is called twice:
- the first time when the action is added to the menu, the second when from the
- menu it is transferred to the toolbar. This is weird since the first widget
- does not ever use the widget so it results in a waste of time.
- */
 QWidget*
 QTMMinibarAction::createWidget (QWidget* parent) {
   static QImage* pxm = xpm_image ("tm_add.xpm"); // See qt_tm_widget.cpp 
@@ -235,7 +230,7 @@ QTMMinibarAction::createWidget (QWidget* parent) {
       tb->setIconSize (sz);
       QFont f = tb->font();
       int fs = as_int (get_preference ("gui:mini-fontsize", QTM_MINI_FONTSIZE));
-      f.setPixelSize (fs > 0 ? fs : QTM_MINI_FONTSIZE);
+      f.setPointSize (qt_zoom (fs > 0 ? fs : QTM_MINI_FONTSIZE));
       tb->setFont(f);
       l->addWidget (tb);
     }
@@ -248,8 +243,6 @@ QTMMinibarAction::createWidget (QWidget* parent) {
  ******************************************************************************/
 
 QTMMenuButton::QTMMenuButton (QWidget* parent) : QToolButton (parent) {
-  QTMAuxMenu m;
-  m.myInitStyleOption (&option);
   setAttribute (Qt::WA_Hover);
 }
 
@@ -274,11 +267,13 @@ QTMMenuButton::mouseReleaseEvent (QMouseEvent* e) {
 void
 QTMMenuButton::paintEvent (QPaintEvent* e) {
   (void) e;
-  QPainter p (this);
   
     // initialize the options
   QStyleOptionToolButton opt;
   initStyleOption (&opt);
+
+  QPainter p (this);
+  QStyleOptionToolButton option;
   QRect r = rect();
   option.rect = r;
   option.state = QStyle::State_Enabled | (opt.state & QStyle::State_MouseOver
@@ -296,14 +291,13 @@ QTMMenuButton::paintEvent (QPaintEvent* e) {
  ******************************************************************************/
 
 QTMMenuWidget::QTMMenuWidget (QWidget* parent) : QWidget (parent) {
-  QTMAuxMenu m;
-  m.myInitStyleOption (&option);
 }
 
 void
 QTMMenuWidget::paintEvent(QPaintEvent* e) {
   QPainter p (this);
-  option.rect = rect ();
+  QStyleOptionMenuItem option;
+  option.rect = rect();
   style()->drawControl (QStyle::CE_MenuEmptyArea, &option, &p, this);
   QWidget::paintEvent (e);
 }
@@ -339,95 +333,65 @@ QTMLazyMenu::attachTo (QAction* a) {
 }
 
 void
-QTMLazyMenu::transferActions (QWidget* from) {
+QTMLazyMenu::transferActions (QList<QAction*>* from) {
   if (from == NULL) return;
   QList<QAction*> list = actions();
   while (!list.isEmpty()) {
     QAction* a = list.takeFirst();
     removeAction (a);
-    a->deleteLater();
   }
-  list = from->actions();
-  while (!list.isEmpty()) {
-    QAction* a = list.takeFirst();
+  while (!from->isEmpty()) {
+    QAction* a = from->takeFirst();
     addAction (a);
-    a->setParent (this);
   }
 }
 
 void
 QTMLazyMenu::force () {
-  QAction* a = concrete (promise_widget())->as_qaction();
-  transferActions (a->menu());
-  delete a;
+BEGIN_SLOT
+  QList<QAction*>* list = concrete (promise_widget())->get_qactionlist();
+  transferActions (list);
+END_SLOT
 }
 
 void
 QTMLazyMenu::destroy (QObject* obj) {
+BEGIN_SLOT
   (void) obj;
   deleteLater();
+END_SLOT
 }
 
-
 /******************************************************************************
  * QTMInputTextWidgetHelper
  ******************************************************************************/
 
-QTMInputTextWidgetHelper::QTMInputTextWidgetHelper (qt_widget _wid,
-                                                    QTMLineEdit* le)
-: QObject (le), p_wid (_wid), done (false) {
+QTMInputTextWidgetHelper::QTMInputTextWidgetHelper (qt_widget _wid, bool _cac): QObject (), p_wid (_wid), can_autocommit (_cac) {
+  QTMLineEdit* le = qobject_cast<QTMLineEdit*>(wid()->qwid);
+  setParent(le);
   ASSERT (le != NULL, "QTMInputTextWidgetHelper: expecting valid QTMLineEdit");
   QObject::connect (le, SIGNAL (returnPressed ()), this, SLOT (commit ()));
   QObject::connect (le, SIGNAL (focusOut (Qt::FocusReason)),
                     this, SLOT (leave (Qt::FocusReason)));
 }
 
-void
-QTMInputTextWidgetHelper::apply () {
-  if (done) return;
-  done = true;
-  the_gui->process_command (wid()->cmd, wid()->ok
-                            ? list_object (object (wid()->input))
-                            : list_object (object (false)));
-}
-
 /*! Executed when the enter key is pressed. */
 void
 QTMInputTextWidgetHelper::commit () {
-  QTMLineEdit* le = qobject_cast <QTMLineEdit*> (sender());
-  if (!le) return;
-
-  done         = false;
-  wid()->ok    = true;
-  wid()->input = from_qstring (le->text());
-
-    // HACK: restore focus to the main editor widget
-  widget_rep* win = qt_window_widget_rep::widget_from_qwidget (le);
-  if (win && concrete (win)->type == qt_widget_rep::texmacs_widget)
-    concrete (get_canvas (win))->qwid->setFocus();
-
-  if (win) apply();    // This is 0 inside a dialog => no command
+BEGIN_SLOT
+  if (sender() != wid()->qwid) return;
+  wid()->commit(true);
+END_SLOT
 }
 
 /*! Executed after commit of the input field (enter) and when losing focus */
 void
 QTMInputTextWidgetHelper::leave (Qt::FocusReason reason) {
-  QTMLineEdit* le = qobject_cast <QTMLineEdit*> (sender());
-  if (!le) return;
-
-    // Don't autocommit if the user pressed the escape key
-    // (see QTMLineEdit::focusOut()).
-  if (reason != Qt::OtherFocusReason &&
-      get_preference ("gui:line-input:autocommit") == "#t") {
-    done         = false;
-    wid()->ok    = true;
-    wid()->input = from_qstring (le->text());
-  } else {
-    le->setText (to_qstring (wid()->input));
-  }
-
-  widget_rep* win = qt_window_widget_rep::widget_from_qwidget (le);
-  if (win) apply();    // This is 0 inside a dialog => no command
+BEGIN_SLOT
+  if (sender() != wid()->qwid) return;
+  wid()->commit((reason != Qt::OtherFocusReason && can_autocommit &&
+                 get_preference ("gui:line-input:autocommit") == "on"));
+END_SLOT
 }
 
 /******************************************************************************
@@ -440,11 +404,19 @@ QTMFieldWidgetHelper::QTMFieldWidgetHelper (qt_widget _wid, QComboBox* cb)
   QObject::connect (cb, SIGNAL (editTextChanged (const QString&)),
                     this, SLOT (commit (const QString&)));
 }
+QTMFieldWidgetHelper::QTMFieldWidgetHelper (qt_widget _wid, QLineEdit* cb)
+: QObject (cb), wid (_wid), done (false) {
+  ASSERT (cb != NULL, "QTMFieldWidgetHelper: expecting valid QLineEdit");
+  QObject::connect (cb, SIGNAL (textChanged (const QString&)),
+                    this, SLOT (commit (const QString&)));
+}
 
 void
 QTMFieldWidgetHelper::commit (const QString& qst) {
+BEGIN_SLOT
   static_cast<qt_field_widget_rep*> (wid.rep)->input =
       scm_quote (from_qstring (qst));
+END_SLOT
 }
 
 /******************************************************************************
@@ -454,25 +426,50 @@ QTMFieldWidgetHelper::commit (const QString& qst) {
 QTMLineEdit::QTMLineEdit (QWidget* parent, string _type, string _ww,
                           int style, command _cmd)
   : QLineEdit (parent), completing (false),
-    type (_type), ww (_ww), cmd (_cmd), last_key (0) {
+    type ("default"), name ("default"), serial ("default"),
+    ww (_ww), cmd (_cmd), last_key (0) {
+  set_type (_type);
+  if (type == "password") setEchoMode(QLineEdit::Password);
   if (style & WIDGET_STYLE_MINI) {
     setStyle (qtmstyle());
       // FIXME: we should remove this and let the scheme code decide.
+#ifdef OS_MACOS
     QPalette pal (palette());
     pal.setColor (QPalette::Base, QColor (252, 252, 248));
     pal.setColor (QPalette::WindowText, Qt::black);
     setPalette (pal);
+#endif
   }
   
-    // just to be sure we don't capture the wrong keys in keyPressEvent
+  // just to be sure we don't capture the wrong keys in keyPressEvent
   setCompleter (0);
 
+  setFocusPolicy (Qt::StrongFocus);
   qt_apply_tm_style (this, style);
 }
 
+void
+QTMLineEdit::set_type (string t) {
+  int i= search_forwards (":", 0, t);
+  if (i >= 0) {
+    type= t (i+1, N(t));
+    name= t (0, i);
+    int j= search_forwards ("#", 0, name);
+    if (j >= 0) {
+      serial= name (j+1, N(name));
+      name  = name (0, j);
+    }
+  }
+  else type= t;
+}
+
 bool
 QTMLineEdit::continuous () {
-  return type == "search" || starts (type, "replace-");
+  return
+    starts (type, "search") ||
+    starts (type, "replace-") ||
+    starts (type, "spell") ||
+    starts (serial, "form-");
 }
 
 /*
@@ -481,19 +478,21 @@ QTMLineEdit::continuous () {
  */
 bool
 QTMLineEdit::event (QEvent* ev) {
-  if (ev->type() == QEvent::KeyPress)  // Handle ALL keys
-    keyPressEvent (static_cast<QKeyEvent*> (ev));
-  else
-    return QWidget::event (ev);
-  return true;
+  if (ev->type() == QEvent::KeyPress) {
+    QKeyEvent *keyEvent= static_cast<QKeyEvent*> (ev);
+    keyPressEvent (keyEvent);
+    return true;
+  }
+  return QLineEdit::event (ev);
 }
 
 extern hashmap<int,string> qtkeymap;
+void initkeymap ();
 
 /*
  FIXME: This is a hideous mess...
  */
-void 
+void
 QTMLineEdit::keyPressEvent (QKeyEvent* ev)
 {
   QCompleter* c = completer();
@@ -503,6 +502,7 @@ QTMLineEdit::keyPressEvent (QKeyEvent* ev)
             : ev->key();
 
   if (continuous ()) {
+    initkeymap ();
     if ((last_key != Qt::Key_Tab || type == "replace-what") &&
         (last_key != Qt::Key_Backtab || type == "replace-by") &&
         last_key != Qt::Key_Down &&
@@ -510,6 +510,8 @@ QTMLineEdit::keyPressEvent (QKeyEvent* ev)
         last_key != Qt::Key_Enter &&
         last_key != Qt::Key_Return &&
         last_key != Qt::Key_Escape &&
+        (!starts (type, "spell") || last_key < 49 || last_key >= 58) &&
+        (!starts (type, "spell") || last_key != 43) &&
         (ev->modifiers() & Qt::ControlModifier) == 0 &&
         (ev->modifiers() & Qt::MetaModifier) == 0)
       QLineEdit::keyPressEvent (ev);
@@ -517,13 +519,13 @@ QTMLineEdit::keyPressEvent (QKeyEvent* ev)
     string s  = from_qstring (text());
     if (last_key >= 32 && last_key <= 126) {
       key= string ((char) last_key);
-      if (key[0] >= 'A' && key[0] <= 'Z')
+      if (is_upcase (key[0]))
         if ((ev->modifiers() & Qt::ShiftModifier) == 0)
           key[0]= (int) (key[0] + ((int) 'a') - ((int) 'A'));
     }
     if (qtkeymap->contains (last_key)) key= qtkeymap[last_key];
     if ((ev->modifiers() & Qt::ShiftModifier) && N(key) > 1) key= "S-" * key;
-#ifdef Q_WS_MAC
+#ifdef Q_OS_MAC
     if (ev->modifiers() & Qt::ControlModifier) key= "C-" * key;
     if (ev->modifiers() & Qt::AltModifier) key= "none";
     if (ev->modifiers() & Qt::MetaModifier) key= "M-" * key;
@@ -695,6 +697,15 @@ QTMLineEdit::keyPressEvent (QKeyEvent* ev)
   }
 }
 
+void
+QTMLineEdit::inputMethodEvent (QInputMethodEvent* ev) {
+  QLineEdit::inputMethodEvent (ev);
+  if (!ev->commitString().isEmpty() && ev->preeditString().isEmpty()) {
+    string str = from_qstring(ev->commitString());
+    cmd (list_object (list_object (object (str), object (str))));
+  }
+}
+
 QSize
 QTMLineEdit::sizeHint () const {
   return qt_decode_length (ww, "", QLineEdit::sizeHint(), fontMetrics());
@@ -737,6 +748,7 @@ QTMTabWidget::QTMTabWidget (QWidget *p) : QTabWidget(p) {
  */
 void
 QTMTabWidget::resizeOthers (int current) {
+BEGIN_SLOT
   for (int i = 0; i < count(); ++i) {
     if (i != current)
       widget(i)->setSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored);
@@ -755,15 +767,15 @@ QTMTabWidget::resizeOthers (int current) {
   if (window()->minimumSize()!=QSize (0,0) && 
       window()->maximumSize() != QSize (QWIDGETSIZE_MAX, QWIDGETSIZE_MAX))
     window()->setFixedSize (window()->sizeHint());
+END_SLOT
 }
 
-
-
 /******************************************************************************
  * QTMRefreshWidget
  ******************************************************************************/
 
 widget make_menu_widget (object wid);
+extern bool menu_caching;
 
 QTMRefreshWidget::QTMRefreshWidget (qt_widget _tmwid, string _strwid, string _kind)
 : QWidget (), strwid (_strwid), kind (_kind),
@@ -796,7 +808,7 @@ QTMRefreshWidget::recompute (string what) {
     object uwid = eval (s);
     cur = make_menu_widget (uwid);
     tmwid->add_child (cur); // FIXME?! Is this ok? what when we refresh?
-    cache (xwid) = cur;
+    if (menu_caching) cache (xwid) = cur;
     return true;
   }
 }
@@ -827,6 +839,7 @@ QTMRefreshWidget::deleteLayout (QLayout* l) {
 
 void
 QTMRefreshWidget::doRefresh (string kind) {
+BEGIN_SLOT
   if (recompute (kind)) {
     if (qwid) qwid->setParent (NULL);
     delete qwid;
@@ -844,6 +857,7 @@ QTMRefreshWidget::doRefresh (string kind) {
         window()->maximumSize() != QSize (QWIDGETSIZE_MAX, QWIDGETSIZE_MAX))
       window()->setFixedSize (window()->sizeHint());  
   }
+END_SLOT
 }
 
 
@@ -904,6 +918,7 @@ QTMRefreshableWidget::deleteLayout (QLayout* l) {
 
 void
 QTMRefreshableWidget::doRefresh (string kind) {
+BEGIN_SLOT
   if (recompute (kind)) {
     if (qwid) qwid->setParent (NULL);
     delete qwid;
@@ -921,6 +936,7 @@ QTMRefreshableWidget::doRefresh (string kind) {
         window()->maximumSize() != QSize (QWIDGETSIZE_MAX, QWIDGETSIZE_MAX))
       window()->setFixedSize (window()->sizeHint());  
   }
+END_SLOT
 }
 
 
@@ -941,7 +957,8 @@ QTMComboBox::QTMComboBox (QWidget* parent) : QComboBox (parent) {
   opt.activeSubControls = QStyle::SC_ComboBoxArrow;
   QRect r = style()->subControlRect (QStyle::CC_ComboBox, &opt,
                                      QStyle::SC_ComboBoxArrow, &cb);
-  minSize.setWidth (r.width());
+  int max_w= (int) floor (40 * retina_scale);
+  minSize.setWidth (min (r.width(), max_w));
 }
 
 /*! Add items and fix the ComboBox size using texmacs length units.
@@ -1006,8 +1023,7 @@ QTMComboBox::event (QEvent* ev) {
  It also scrolls the viewport to the position of selected items in QListWidgets.
  */
 void
-QTMScrollArea::setWidgetAndConnect (QWidget* w)
-{
+QTMScrollArea::setWidgetAndConnect (QWidget* w) {
   setWidget (w);
  
   listViews = w->findChildren<QTMListView*>();
@@ -1020,8 +1036,8 @@ QTMScrollArea::setWidgetAndConnect (QWidget* w)
 
 /*! Scrolls the area to a given index in a QTMListView. */
 void
-QTMScrollArea::scrollToSelection (const QItemSelection& sel)
-{
+QTMScrollArea::scrollToSelection (const QItemSelection& sel) {
+BEGIN_SLOT
   if (sel.isEmpty())
     return;
 
@@ -1035,6 +1051,7 @@ QTMScrollArea::scrollToSelection (const QItemSelection& sel)
     if (! viewport()->geometry().contains (x, y))
       ensureVisible (x, y, r.width(), r.height());
   }
+END_SLOT
 }
 
 /*! Work around a problem with scrolling before the widget is shown.
@@ -1043,8 +1060,7 @@ QTMScrollArea::scrollToSelection (const QItemSelection& sel)
  insufficient amount. See the comments to QTMScrollArea.
  */
 void
-QTMScrollArea::showEvent (QShowEvent* ev)
-{
+QTMScrollArea::showEvent (QShowEvent* ev) {
   for (ListViewsIterator it = listViews.begin(); it != listViews.end(); ++it) {
     QItemSelection sel = (*it)->selectionModel()->selection();
     (*it)->selectionChanged (sel, sel);
@@ -1052,7 +1068,6 @@ QTMScrollArea::showEvent (QShowEvent* ev)
   QScrollArea::showEvent (ev);
 }
 
-
 /******************************************************************************
  * QTMListView
  ******************************************************************************/
@@ -1078,12 +1093,18 @@ QTMListView::QTMListView (const command& cmd,
   setSelectionMode (multiple ? ExtendedSelection : SingleSelection);
   setEditTriggers (NoEditTriggers);
 
+    // NOTE: using selectionModel()->select(item, QItemSelection::SelectCurrent)
+    // doesn't update the selection but overwrites it, so we explicitly define
+    // our QItemSelection and use merge()
+  QItemSelection sel;
   for (int i = 0; i < model()->rowCount(); ++i) {
     QModelIndex item = model()->index (i, 0);
-    if (selections.contains (model()->data (item, Qt::DisplayRole).toString(), Qt::CaseSensitive))
-      selectionModel()->select (item, QItemSelectionModel::SelectCurrent);
+    if (selections.contains (model()->data (item, Qt::DisplayRole).toString(),
+                             Qt::CaseSensitive))
+      sel.merge (QItemSelection(item, item), QItemSelectionModel::Select);
   }
-
+  selectionModel()->select (sel, QItemSelectionModel::Select);
+  
   if (!scroll) {
     setMinimumWidth (sizeHintForColumn(0));
     setMinimumHeight (sizeHintForRow(0) * model()->rowCount());
@@ -1110,11 +1131,12 @@ QTMListView::QTMListView (const command& cmd,
  */
 void
 QTMListView::selectionChanged (const QItemSelection& c, const QItemSelection& p) {
+BEGIN_SLOT
   QListView::selectionChanged (c, p);
   emit selectionChanged (c);
+END_SLOT
 }
 
-
 /******************************************************************************
  * QTMTreeView
  ******************************************************************************/
@@ -1125,12 +1147,22 @@ QTMTreeView::QTMTreeView (command cmd, tree data, const tree& roles, QWidget* p)
   setUniformRowHeights (true);  // assuming we display only text.
   setHeaderHidden (true);       // for now...
   QObject::connect (this, SIGNAL (pressed (const QModelIndex&)),
-                    this,   SLOT (callOnClick (const QModelIndex&)));
+                    this,   SLOT (callOnChange (const QModelIndex&)));
 }
 
 void
-QTMTreeView::callOnClick (const QModelIndex& index) {
-  object arguments = list_object ((int)QApplication::mouseButtons());
+QTMTreeView::currentChanged (const QModelIndex& curr, const QModelIndex& prev) {
+  (void) prev;
+  if (selectedIndexes().contains(curr))
+    callOnChange (curr, false);
+}
+
+void
+QTMTreeView::callOnChange (const QModelIndex& index, bool mouse) {
+BEGIN_SLOT
+  object arguments = mouse ? list_object ((int)QApplication::mouseButtons())
+                           : list_object (-1);
+    
     // docs state the index is valid, no need to check
   QVariant d = tmModel()->data (index, QTMTreeModel::CommandRole);
     // If there's no CommandRole, we return the subtree by default
@@ -1145,6 +1177,7 @@ QTMTreeView::callOnClick (const QModelIndex& index) {
     d = tmModel()->data (index, ++cnt);
   }
   _cmd (arguments);
+END_SLOT
 }
 
 inline QTMTreeModel*
