@@ -9,23 +9,28 @@
  * in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.
  ******************************************************************************/
 
-#include "QTMStyle.hpp"
-#include <QApplication>
-#include <QStyleOptionMenuItem>
 #include "tm_ostream.hpp"
+#include "renderer.hpp"
+
+#include "QTMStyle.hpp"
+
+#include <QStyleOptionMenuItem>
 #include <qdrawutil.h>
 #include <QPainter>
+#include <QPainterPath>
 #include <QMainWindow>
+#include <QApplication>
+
+int
+qt_zoom (int sz) {
+  return (int) (retina_scale * ((double) sz));
+}
 
 QStyle*
 qtmstyle () {
-  static QStyle* qtmstyle= NULL;
-  if (!qtmstyle) {
-    qtmstyle = new QTMStyle ();
-  }
-  if (!qtmstyle) {
-    qtmstyle = qApp->style ();
-  }
+  static QStyle* qtmstyle = NULL;
+  if (!qtmstyle)
+    qtmstyle = new QTMStyle (qApp->style());
   return qtmstyle;
 }
 
@@ -33,17 +38,22 @@ qtmstyle () {
  * QTMProxyStyle (does not own *style)
  ******************************************************************************/
 
-QTMProxyStyle::QTMProxyStyle (QStyle* _base):
-QStyle (), base (_base) {}
-
-QTMProxyStyle::~QTMProxyStyle() {
-  // delete style;
-}
+QTMProxyStyle::QTMProxyStyle (QStyle* _base) : QStyle (), base (_base) { }
 
 inline  QStyle *QTMProxyStyle::baseStyle() const {
   return ( base ? base : qApp->style() );
 }
 
+#if (QT_VERSION >= 0x050000)
+int
+QTMProxyStyle::layoutSpacing (QSizePolicy::ControlType control1,
+                              QSizePolicy::ControlType control2,
+                              Qt::Orientation orientation,
+                              const QStyleOption *option,
+                              const QWidget *widget) const {
+  return baseStyle()->layoutSpacing(control1, control2, orientation, option, widget);
+}
+#endif
 
 void
 QTMProxyStyle::drawComplexControl (ComplexControl control, const QStyleOptionComplex* option, QPainter* painter, const QWidget* widget) const {
@@ -194,7 +204,7 @@ static void qtmDrawRoundedRect(QPainter *p, const QRectF &rect, qreal xRadius, q
     qreal rxx2 = w*xRadius/100;
     qreal ryy2 = h*yRadius/100;
     
-#ifdef Q_WS_X11
+#ifdef Q_OS_LINUX
     // There is a bug (probably in arcTo) for small sizes.
     // We use a rough linear approx.
     rxx2 /= 4;
@@ -415,9 +425,11 @@ QTMStyle::drawControl (ControlElement element, const QStyleOption* option, QPain
       }
 #endif
 
+#if (QT_VERSION < 0x050000)
     case CE_ToolBar: {
 #ifdef UNIFIED_TOOLBAR
-      if ((widget) &&  (widget->windowTitle() == "mode toolbar"))  {
+      if (use_unified_toolbar &&
+          (widget) && (widget->windowTitle() == "mode toolbar")) {
 
           // For unified tool bars, draw nothing.
           if (QMainWindow * mainWindow = qobject_cast<QMainWindow *>(widget->window())) {
@@ -453,6 +465,26 @@ QTMStyle::drawControl (ControlElement element, const QStyleOption* option, QPain
         }
 #endif // UNIFIED_TOOLBAR
     } break;
+#else
+  case CE_ToolBar:
+#ifdef UNIFIED_TOOLBAR
+    if (use_unified_toolbar &&
+        (widget) && (widget->windowTitle() == "main toolbar"))
+      break;
+#endif // UNIFIED_TOOLBAR
+
+    if ((widget) && (widget->windowTitle() == "main toolbar"))
+      painter->fillRect(option->rect, QColor (208, 208, 208));
+      //painter->fillRect(option->rect, QColor (192, 192, 192));
+    else if ((widget) && (widget->windowTitle() == "mode toolbar"))
+      painter->fillRect(option->rect, QColor (224, 224, 224));
+      //painter->fillRect(option->rect, QColor (216, 216, 216));
+    else if ((widget) && (widget->windowTitle() == "focus toolbar"))
+      painter->fillRect(option->rect, QColor (240, 240, 240));
+    else 
+      baseStyle()->drawControl (element, option, painter, widget);
+    break;
+#endif
       
     default:
       baseStyle()->drawControl (element, option, painter, widget);
