@@ -22,8 +22,8 @@ box graphics_box (path ip, array<box> bs, frame f, grid g,
 box graphics_group_box (path ip, array<box> bs);
 box text_at_box (path ip, box b, SI x, SI y, SI axis, SI pad);
 box point_box (path ip, point p, SI r, pencil pen, brush br, string style);
-box curve_box (path ip, curve c, pencil pen,
-	       array<bool> style, SI style_unit,
+box curve_box (path ip, curve c, double portion, pencil pen,
+	       array<bool> style, array<point> motif, SI style_unit,
 	       brush fill_br, array<box> arrows);
 box spacial_box (path ip, spacial obj);
 box grid_box (path ip, grid g, frame f, SI un, point lim1, point lim2);
